# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Web Traversal Library (WTL)
Provides a bottom abstraction layer for automatic web workflows.
"""

from .actions import Action, Actions, ElementAction
from .classifiers import ActiveElementFilter, ElementClassifier, ScalingMode, ViewClassifier
from .color import Color
from .config import Config
from .error import ElementNotFoundError, Error, ScrapingError, WebDriverSendError, WindowClosedError
from .geometry import Point, Rectangle
from .javascript import JavascriptWrapper
from .policies import multi_tab_coroutine, single_tab, single_tab_coroutine
from .scraper import Scraper
from .selector import Selector
from .snapshot import Elements, PageElement, PageSnapshot
from .version import __version__
from .view import View
from .window import Window
from .workflow import Workflow



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Configuration object for discovery workflow"""
from __future__ import annotations

import functools
import json
import logging
import os
import re
from pathlib import Path
from typing import Any, Dict, List, Union

from prodict import Prodict  # pylint: disable=syntax-error

from .version_check import VERSION_CMD, run_cmd

logger = logging.getLogger("wtl")


class Config:
    """Represents a config object."""

    CONFIG_DIR = Path(os.path.abspath(os.path.dirname(__file__))) / "configs"

    REQUIRED_PARAMS = dict(
        [
            ("bs_html_parser", str),
            ("timeout", int),
            ("actions.abort.close", bool),
            ("scraping.disable_animations", bool),
            ("scraping.attempts", int),
            ("scraping.prescroll", bool),
            ("scraping.page_load_timeout", int),
            ("scraping.wait_loading", float),
            ("scraping.wait_scroll", float),
            ("scraping.wait_action", float),
            ("scraping.save_mhtml", bool),
            ("scraping.temp_path", str),
            ("scraping.mhtml_timeout", int),
            ("scraping.history", bool),
            ("scraping.full_history", bool),
            ("scrolling.max_page_height", int),
            ("browser.browser", str),
            ("browser.useragent", str),
            ("browser.width", int),
            ("browser.height", int),
            ("browser.headless", bool),
            ("browser.enable_mhtml", bool),
            ("browser.proxy", str),
            ("javascript.info", str),
            ("javascript.warning", str),
            ("javascript.severe", str),
            ("debug.autoscroll", bool),
            ("debug.default_canvas_viewport", bool),
            ("debug.live", bool),
            ("debug.live_delay", float),
            ("debug.live_annotation", bool),
            ("debug.screenshots", bool),
            ("debug.save", bool),
            ("debug.preserve_window", bool),
            ("debug.action_highlight_color", str),
        ]
    )

    def __init__(self, cfg: List[Union[str, Path, dict]] = None):
        if cfg is None:
            cfg = []

        self._instance = Prodict()
        for item in cfg:
            if item:
                self.update(item)

        self._ensure_has_all_params(Config.REQUIRED_PARAMS)
        logger.debug(f"Configuration loaded from {cfg}")

    def update(self, cfg: Union[str, Path, dict]):
        ddict = Config._parse_input_config(cfg)
        Config._update(self._instance, ddict)

    @staticmethod
    def _update(fst_dd, snd_dd):
        for key, snd_value in snd_dd.items():
            fst_value = fst_dd.get(key, None)
            do_recurse = isinstance(fst_value, dict) and isinstance(snd_value, dict)
            if key not in fst_dd or not do_recurse:
                if key in fst_dd:
                    logger.debug(f"Overwriting {key} -> {fst_value} with {snd_value}")
                else:
                    logger.debug(f"Adding {key} -> {snd_dd}")
                fst_dd[key] = snd_value
            elif do_recurse:
                Config._update(fst_value, snd_value)

    def to_json(self):
        return json.loads(json.dumps(self._instance))

    def __getattr__(self, item):
        return getattr(self._instance, item)

    def __getitem__(self, item):
        return functools.reduce(lambda obj, key: obj[key], item.split("."), self._instance)

    def __contains__(self, item):
        obj = self._instance
        for attr in item.split("."):
            if attr not in obj:
                return False
            obj = obj[attr]
        return True

    def _ensure_has_all_params(self, params):
        for param_name, param_type in params.items():
            if param_name not in self:
                raise ValueError(f"Missing required parameter '{param_name}'")
            if not isinstance(self[param_name], param_type):
                raise ValueError(
                    f"Expected parameter '{param_name}' to be of type '{param_type}',"
                    f"got '{type(self[param_name])}' instead"
                )

    def __repr__(self):
        return repr(self._instance)

    def __str__(self):
        return str(self._instance)

    def __len__(self):
        return len(self._instance)

    def validate(self):
        """
        Performs basic sanity checks on the configuration values.
        Throws AssertionError if something is incorrect.
        """
        LOG_LEVELS = ["", "debug", "info", "warning", "error", "critical"]
        BROWSERS = ["chrome", "firefox"]
        cfg = self._instance

        assert cfg.javascript.info in LOG_LEVELS
        assert cfg.javascript.warning in LOG_LEVELS
        assert cfg.javascript.severe in LOG_LEVELS
        assert cfg.timeout >= 0
        assert cfg.scraping.attempts >= 1
        assert cfg.scraping.page_load_timeout >= 0
        assert cfg.scrolling.max_page_height >= 0
        assert cfg.browser.width >= 1
        assert cfg.browser.height >= 1
        assert cfg.debug.live_delay >= 0
        assert cfg.browser.browser in BROWSERS

        if cfg.browser.browser == "chrome":
            assert run_cmd(VERSION_CMD.CHROME) or run_cmd(VERSION_CMD.CHROMIUM)
            assert run_cmd(VERSION_CMD.CHROMEDRIVER)
        elif cfg.browser.browser == "firefox":
            assert run_cmd(VERSION_CMD.FIREFOX)
            assert run_cmd(VERSION_CMD.GECKODRIVER)

    @staticmethod
    def default(cfg: List[Union[str, Path, dict]] = None) -> Config:
        """Creates a Config object based on all default values"""
        cfg = cfg or []
        return Config([Path("default.json")] + cfg)  # type: ignore

    @staticmethod
    def _parse_input_config(cfg: Union[str, Path, dict]):
        if isinstance(cfg, dict):
            return Prodict.from_dict(cfg)

        if isinstance(cfg, str) and "=" in cfg:
            key, value = cfg.split("=")
            parsed_value: Any

            if value == "True":
                parsed_value = True
            elif value == "False":
                parsed_value = False
            elif re.match(r"^[-+]?[0-9]+$", value):
                parsed_value = int(value)
            elif re.match(r"^[-+]?[0-9]*\.?[0-9]+$", value):
                parsed_value = float(value)
            else:
                parsed_value = value

            key_chain = key.split(".")
            result: Dict[str, Any] = {}
            ref = result
            while len(key_chain) > 1:
                next_key = key_chain.pop(0)
                ref[next_key] = {}
                ref = ref[next_key]
            ref[key_chain.pop(0)] = parsed_value
            return Prodict.from_dict(result)

        if isinstance(cfg, str):
            cfg = Path(cfg)

        if isinstance(cfg, Path):
            if not cfg.suffix:
                cfg = cfg.with_suffix(".json")
            if not cfg.exists():
                cfg = Config.CONFIG_DIR / cfg
            if not cfg.exists():
                raise FileNotFoundError(f"Given config {str(cfg)} does not exist locally or as part of wtl.")

            json_data = json.load(Path(cfg).open())
            return Prodict.from_dict(json_data)

        raise ValueError(f"Unexpected config input: {cfg}")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Static snapshot of a web page. Contains source, screenshots, element images, and metadata.
"""
from __future__ import annotations

import json
import logging
import os
import re
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Union

import bs4
from PIL import Image

from .config import Config
from .error import ScrapingError
from .geometry import Point, Rectangle
from .graphics import crop_image
from .processtools import cached_property
from .screenshot import Screenshot
from .selector import Selector

logger = logging.getLogger("wtl")


@dataclass(frozen=True)
class PageElement:
    """
    Represents an element with associated and structured `metadata` on a given `page`.
    """

    page: "PageSnapshot" = field(repr=False)
    metadata: dict = field(repr=False)

    @property
    def raw_scores(self) -> Dict[str, float]:
        """Returns all raw classifier scores."""
        if "raw_scores" not in self.metadata:
            self.metadata["raw_scores"] = {}
        return self.metadata["raw_scores"]

    @cached_property
    def tag(self) -> bs4.Tag:
        """Returns the bs4.Tag associated with this PageElement"""
        tag = self.page.page_source.find(attrs={"wtl-uid": self.wtl_uid})
        if not tag:
            logger.warning(f"No bs4.tag with wtl-uid={self.wtl_uid}!")
        return tag

    @cached_property
    def wtl_uid(self) -> int:
        """Returns the wtl-uid associated with this PageElement."""
        return self.metadata["wtl_uid"]

    @cached_property
    def wtl_parent_uid(self) -> int:
        """Returns the wtl-uid associated with the parent of this PageElement."""
        return self.metadata["wtl_parent_uid"]

    @cached_property
    def parent(self) -> PageElement:
        """Returns the parent of this PageElement."""
        return self.page.elements.by_uid(self.wtl_parent_uid)

    @cached_property
    def location(self) -> Point:
        """Returns the top-left position of this PageElement."""
        location = self.metadata["location"]
        return Point(location["x"], location["y"])

    @cached_property
    def size(self) -> Point:
        """Returns the wtl-uid associated with this PageElement."""
        size = self.metadata["size"]
        return Point(size["width"], size["height"])

    @cached_property
    def bounds(self) -> Rectangle:
        """Returns the bounding box of this PageElement."""
        return Rectangle(self.location, self.location + self.size)

    @cached_property
    def selector(self) -> Selector:
        """CSS Selector for the element without attributes."""
        return Selector.build(self.page.page_source, self.wtl_uid)

    @cached_property
    def font_size(self) -> float:
        """Returns resolved font size property in pixels."""
        return PageElement.parse_resolved_size(self.metadata["font_size"])

    @cached_property
    def screenshot(self) -> Image.Image:
        """
        Returns element screenshot for the given element, cropped from the page screenshot.

        .. warning::
            Screenshotting must have been enabled (and run) on the page for this to work!
        """
        fixed_pos = self.metadata["fixed_pos"]
        page_screenshot = self.page.screenshots["first"] if fixed_pos else self.page.screenshots["full"]
        assert page_screenshot, "Page screenshotting must be enabled if requesting element screenshot"

        screenshot_box = Rectangle(Point.zero(), page_screenshot.size)
        element_box = self.bounds
        intersection_box = Rectangle.intersection([screenshot_box, element_box])

        if intersection_box.area == 0:
            return None

        return crop_image(page_screenshot.image, intersection_box)

    @staticmethod
    def parse_resolved_size(value: str) -> float:
        """
        Parse resolved CSS size value
        :raises: ``ValueError`` if format not supported (currently float followed by ``px``)
        """
        pattern = re.compile(r"\d+(.\d*)?px")
        value = value.strip()
        if not pattern.fullmatch(value):
            raise ScrapingError(f"Expected a number followed by px, got '{value}'")
        return float(value[:-2])


class Elements(list):
    """Helper class for a list of elements from the same page"""

    def by_score(self, name: Union[str, Iterable[str]], limit: float = 0.0) -> Elements:
        """Return elements tagged with all given scores over a certain limit"""
        if name == "all":
            return self

        if isinstance(name, str):
            return Elements([e for e in self if name in e.metadata and e.metadata[name] > limit])

        return Elements([e for e in self if set(name) <= set(e.metadata.keys())])

    def by_raw_score(self, name: Union[str, Iterable[str]], limit: float = 0.0) -> Elements:
        """Return elements tagged with all given raw scores over a certain limit"""
        if name == "all":
            return self

        if isinstance(name, str):
            return Elements([e for e in self if name in e.raw_scores and e.raw_scores[name] > limit])

        return Elements([e for e in self if set(name) <= set(e.metadata.keys())])

    def by_selector(self, selector: Selector) -> Elements:
        """Return all elements that match a given selector"""
        if not self:
            return Elements([])

        tags = set(self[0].page.page_source.html.select(selector.css))
        if not tags:
            return Elements([])

        wtl_uids = set(int(x.attrs["wtl-uid"]) for x in tags if "wtl-uid" in x.attrs)
        elements = Elements([e for e in self if e.wtl_uid in wtl_uids])

        # Falls back on BS4 tags if selector matches something that hasn't been snapshotted yet
        if not elements:
            elements = Elements([e for e in self if e.tag in tags])

        return elements

    def by_subtree(self, target: Union[Selector, PageElement], include_root: bool = True) -> Elements:
        """Return all elements belonging to the subtree where given element is the root"""
        if not self:
            return Elements([])

        if isinstance(target, Selector):
            targets = self.by_selector(target)
            assert len(targets) == 1
            target = targets[0]

        assert isinstance(target, PageElement)
        tags = set(target.tag.descendants)
        return Elements([e for e in self if e.tag in tags] + ([target] if include_root else []))

    def by_uid(self, wtl_uid: int) -> PageElement:
        """Returns the element with the given wtl_uid"""
        for e in self:
            if e.wtl_uid == wtl_uid:
                return e
        return None

    def sort_by(self, name: str = None, reverse: bool = False) -> Elements:
        """
        Sorts by a certain (raw) score. If given name does not exist the element gets (raw) score 0.
        """
        self.sort(key=lambda e: e.raw_scores.get(name, 0), reverse=reverse)
        return self


@dataclass(repr=False, frozen=True)
class PageSnapshot:
    """
    Static snapshot of a web page.
    Contains source (DOM), screenshots, elements and metadata.

    .. note::
        Not _all_ parts of the website (separate CSS and JS files, for example) are
        represented here. If that's what you need, consider storing MHTML snapshots
        and manipulating them manually.
    """

    page_source: bs4.BeautifulSoup
    page_metadata: Optional[dict]
    elements_metadata: List[dict]
    elements: Elements = field(init=False)
    screenshots: Dict[str, Screenshot] = field(default_factory=dict)
    mhtml_source: bytes = None

    def __post_init__(self):
        page_elements = Elements([PageElement(self, metadata) for metadata in self.elements_metadata])
        object.__setattr__(self, "elements", page_elements)

        if "screenshots" not in self.page_metadata:
            self.page_metadata["screenshots"] = []

    def new_screenshot(self, name: str, of: str) -> Screenshot:
        """
        Creates a new screenshot from a copy of a previous one.
        Saves it to this snapshot and returns the new screenshot object.
        """
        assert of in self.screenshots.keys(), "Creating screenshot from nonexistant original!"
        self.screenshots[name] = self.screenshots[of].copy(name)
        return self.screenshots[name]

    @classmethod
    def load(cls, path: Path, bs4_parser: str = None) -> PageSnapshot:
        """
        Returns a PageSnapshot at given path (directory) in the format
        as stored in :func:`save`.
        """
        assert path.exists()
        screenshots: Dict[str, Screenshot] = {}
        mhtml_source: bytes = None

        if bs4_parser is None:
            bs4_parser = Config.default().bs_html_parser

        with open(path / "source.html") as f:
            page_source = bs4.BeautifulSoup(f.read(), bs4_parser)
        with open(path / "page_metadata.json") as f:
            page_metadata = json.load(f)
        with open(path / "elements_metadata.json") as f:
            elements_metadata = json.load(f)
        for f in page_metadata["screenshots"]:
            screenshots[str(f)] = Screenshot.load(str(f), path / f"{f}.png")
        try:
            with open(path / "page.mhtml", "rb") as f:  # type: ignore
                mhtml_source = f.read()  # type: ignore
        except OSError:
            pass

        snapshot = PageSnapshot(
            page_source=page_source,
            page_metadata=page_metadata,
            elements_metadata=elements_metadata,
            screenshots=screenshots,
            mhtml_source=mhtml_source,
        )

        return snapshot

    def save(self, path: Path):
        """
        Saves the current PageSnapshot instance to a folder.
        The number of files depends on whether screenshots were taken or not.
        """
        os.makedirs(path, exist_ok=True)

        with open(path / "source.html", "w") as f:
            f.write(str(self.page_source))
        with open(path / "page_metadata.json", "w") as f:
            json.dump(self.page_metadata, f)
        with open(path / "elements_metadata.json", "w") as f:
            json.dump(self.elements_metadata, f)
        for scr in self.screenshots.values():
            scr.save(path)
        if self.mhtml_source:
            with open(path / "page.mhtml", "wb") as f:  # type: ignore
                f.write(self.mhtml_source)  # type: ignore



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Module for different webdrivers considered.
"""
import importlib.resources
import json
import logging
import os.path
from pathlib import Path
from typing import List

from selenium import webdriver
from selenium.webdriver import DesiredCapabilities
from selenium.webdriver.remote.webdriver import WebDriver

from .config import Config
from .error import WebDriverSendError
from .version_check import VERSION_CMD, run_cmd

logger = logging.getLogger("wtl")


def _setup_chrome(config: Config, profile_path: Path = None) -> WebDriver:
    assert run_cmd(VERSION_CMD.CHROME, "Chrome version") or run_cmd(VERSION_CMD.CHROMIUM, "Chromium version")
    assert run_cmd(VERSION_CMD.CHROMEDRIVER, "Chromedriver version")

    chrome_options = webdriver.ChromeOptions()
    browser = config.browser

    # Disable security
    # Don't enforce the same-origin policy, see https://developer.mozilla.org/en-US/docs/Web/HTTP/CORS
    # Without this the JS code executed with Selenium WebDriver cannot access cross-domain CSS which are popular.
    chrome_options.add_argument("--disable-web-security")
    chrome_options.add_argument("--allow-running-insecure-content")
    chrome_options.add_argument("--no-sandbox")

    # Disable unused/annoying features
    chrome_options.add_argument("--disable-infobars")
    chrome_options.add_argument("--mute-audio")
    chrome_options.add_argument("--hide-scrollbars")
    chrome_options.add_argument("--disable-dev-shm-usage")

    chrome_options.headless = browser.headless

    if profile_path:
        profile_dir = profile_path.name
        user_data_dir = str(profile_path.parent)
        chrome_options.add_argument(f"--user-data-dir={user_data_dir}")
        chrome_options.add_argument(f"--profile-directory={profile_dir}")

    if browser.proxy:
        chrome_options.add_argument(f"--proxy-server={browser.proxy}")
        chrome_options.add_argument("--disk-cache-dir=chrometemp/")  # Proxy messes up access rights of cache folder

    if browser.enable_mhtml:
        with importlib.resources.path("webtraversallibrary.extension", "wtl_extension.crx") as filepath:
            chrome_options.add_extension(str(filepath))
        chrome_options.add_experimental_option(
            "prefs", {"download.default_directory": os.path.expanduser(config.scraping.temp_path)}
        )

    if "pixelRatio" in browser and browser.pixelRatio != 0:
        # Emulating a mobile device - the full set of parameters must be provided
        chrome_options.add_experimental_option(
            "mobileEmulation",
            {
                "deviceMetrics": {"width": browser.width, "height": browser.height, "pixelRatio": browser.pixelRatio},
                "userAgent": browser.useragent,
            },
        )
    else:
        if "useragent" in browser:
            chrome_options.add_argument(f'--user-agent="{browser.useragent}"')

    if "width" in browser and "height" in browser:
        chrome_options.add_argument(f"--window-size={browser.width},{browser.height + 120}")

    capabilities = DesiredCapabilities.CHROME.copy()
    capabilities["goog:loggingPrefs"] = {"browser": "ALL"}

    def _add_script(driver: WebDriver, script_path: str):
        send(driver, "Page.addScriptToEvaluateOnNewDocument", {"source": script_path})

    WebDriver.add_script = _add_script

    # launch Chrome
    driver = webdriver.Chrome(options=chrome_options, desired_capabilities=capabilities)

    return driver


def _setup_firefox(config: Config, _profile_path: Path = None) -> WebDriver:
    assert run_cmd(VERSION_CMD.FIREFOX, "Firefox version")
    assert run_cmd(VERSION_CMD.GECKODRIVER, "Geckodriver version")

    firefox_profile = webdriver.FirefoxProfile()
    firefox_options = webdriver.FirefoxOptions()
    browser = config.browser

    # Disable security
    firefox_options.add_argument("-safe-mode")
    firefox_profile.set_preference("browser.link.open_newwindow", 3)
    firefox_profile.set_preference("browser.link.open_newwindow.restriction", 0)
    firefox_profile.set_preference("media.volume_scale", "0.0")

    firefox_options.headless = browser.headless

    if browser.proxy:
        address = browser.proxy.split(":")
        firefox_profile.set_preference("network.proxy.http", address[0])
        firefox_profile.set_preference("network.proxy.http_port", address[1])

    if browser.enable_mhtml:
        # TODO
        logger.critical("Firefox does not currently support MHTML")
        assert False

    if "pixelRatio" in browser and browser.pixelRatio != 0:
        # TODO
        logger.error("Firefox does not currently support custom pixelratio")

    if "useragent" in browser:
        firefox_profile.set_preference("general.useragent.override", browser.useragent)

    def _add_script(driver: WebDriver, script_path: str):
        send(driver, "Page.addScriptToEvaluateOnNewDocument", {"source": script_path})

    WebDriver.add_script = _add_script

    # TODO No support to output console logs for geckodriver

    driver = webdriver.Firefox(options=firefox_options, firefox_profile=firefox_profile)

    # Screen size (we can't set viewport size as for Chrome so we adjust screen size)
    if "width" in browser and "height" in browser:
        driver.set_window_size(browser.width, browser.height)

    return driver


def setup_driver(config: Config, profile_path: Path = None, preload_callbacks: List[Path] = None) -> WebDriver:
    """
    Creates a WebDriver object with the given configuration.
    :param: configuration Configuration dictionary
    :return: A WebDriver instance
    """
    driver = None
    name = config.browser.browser.lower()

    if name == "chrome":
        driver = _setup_chrome(config=config, profile_path=profile_path)
    elif name == "firefox":
        driver = _setup_firefox(config=config, _profile_path=profile_path)
    else:
        raise NotImplementedError(f"Uninmplemented browser type given to setup_driver: {config.browser.browser}")

    preload_callbacks = preload_callbacks or []
    for cb in preload_callbacks:
        driver.add_script(str(cb))
    return driver


def send(driver: WebDriver, cmd: str, params: dict = None) -> int:
    """
    Send command to the webdriver, return resulting status code.
    """
    # pylint: disable=protected-access
    if params is None:
        params = {}
    resource = f"/session/{driver.session_id}/chromium/send_command_and_get_result"
    url = driver.command_executor._url + resource
    body = json.dumps({"cmd": cmd, "params": params})
    response = driver.command_executor._request("POST", url, body)
    value = response.get("value")

    if response.get("status", False):
        raise WebDriverSendError(f"Command '{cmd}' returned status={value}")

    return value



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Helper function for logging.
"""

import logging
import pathlib
from typing import Optional


def setup_logging(log_dir: Optional[pathlib.Path] = None, logging_level: int = logging.INFO):
    """
    Sets up logging: create a directory to write log files to, configure handlers. Sets sane
    default values for in-house and third-party modules.

    Will remove any existing logging handlers with the name "webtraversallibrary" before proceeding.

    :param log_dir: directory to write log files to.
    :param logging_level: level of logging you wish to have, accepts number or logging.LEVEL
    """
    logging.getLogger("selenium").setLevel(logging.INFO)
    logging.getLogger("PIL").setLevel(logging.WARNING)
    logging.getLogger("urllib3").setLevel(logging.WARNING)

    logger = logging.getLogger()

    for h in logger.handlers[:]:
        if h.name and "webtraversallibrary" in h.name:
            logger.removeHandler(h)

    formatter = logging.Formatter("{asctime} {levelname:8} {name:12} - {message}", "%H:%M:%S", style="{")
    to_console = logging.StreamHandler()
    to_console.setFormatter(formatter)
    to_console.set_name("webtraversallibrary-console")  # type: ignore
    logger.setLevel(logging_level)
    logger.addHandler(to_console)

    if log_dir:
        # If log directory is provided, write to there as well
        log_dir.mkdir(exist_ok=True, parents=True)
        to_file = logging.FileHandler(log_dir / "output.log", mode="w", encoding="utf-8")
        to_file.setFormatter(formatter)
        to_file.set_name("webtraversallibrary-file")  # type: ignore
        logger.addHandler(to_file)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Module collecting helper functions for common processing tasks, such as muting stdout within a context.
"""

import logging
import signal

logger = logging.getLogger("wtl")


class cached_property:
    """
    Decorator that caches a property return value and will return it on later calls.
    Adapted from The Python Cookbok, 2nd edition.

    .. note::
        If you want to map different arguments to values, use `functools.lru_cache`!
    """

    def __init__(self, method):
        self.method = method
        self.name = method.__name__
        self.__doc__ = method.__doc__

    def __get__(self, inst, cls):
        if inst is None:
            return self
        result = self.method(inst)
        object.__setattr__(inst, self.name, result)
        return result


class TimeoutContext:
    """
    Uses :mod:`signal` to raise TimeoutError within the block, if execution went over a specified timeout.
    """

    def __init__(self, n_seconds, error_class=TimeoutError):
        self.n_seconds = n_seconds
        self.error_cls = error_class

    def __enter__(self):
        if self.n_seconds > 0:
            signal.signal(signal.SIGALRM, self.raise_error)
            signal.alarm(self.n_seconds)
            logger.debug(f"TimeoutContext: Operation timeout is set to {self.n_seconds} sec.")

    def __exit__(self, *args, **kwargs):
        # Cancel the alarm
        if self.n_seconds > 0:
            signal.signal(signal.SIGALRM, signal.SIG_DFL)
            time_left = signal.alarm(0)
            if time_left > 0:
                logger.debug(f"TimeoutContext: Operation finished without triggering the alarm ({time_left} sec. left)")
            else:
                logger.debug("TimeoutContext: Operation was interrupted by the timeout")

    def raise_error(self, signal_num, _):
        assert signal_num == signal.SIGALRM
        raise self.error_cls(f"Operation timed out after {self.n_seconds} sec.")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Wrapper functions around JavaScript code to be used from Selenium WebDriver. The main reason to store JS code in
files instead of embedding it in Python code is convenience: it is more readable and has better IDE support.
"""

import functools
import logging
import os
from datetime import datetime
from functools import wraps
from pathlib import Path
from typing import Any, Dict, Iterable, List, Union

from selenium.common.exceptions import (
    JavascriptException,
    NoAlertPresentException,
    UnexpectedAlertPresentException,
    WebDriverException,
)
from selenium.webdriver.common.alert import Alert
from selenium.webdriver.remote.webdriver import WebDriver

from .color import Color
from .config import Config
from .error import WindowClosedError
from .geometry import Point, Rectangle
from .selector import Selector

logger = logging.getLogger("wtl")


def safe_selenium_method(func):
    """
    Handles errors thrown in the browser while executing javascript and outputs information to the log.
    Note: This is a clumsy decorator for instance methods and assumes there is a self.driver member.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        def to_logger(msg, js_level):
            if js_level == "INFO":
                level = self.config.javascript.info
            elif js_level == "WARNING":
                level = self.config.javascript.warning
            elif js_level == "SEVERE":
                level = self.config.javascript.severe
            else:
                return
            if level == "debug":
                logger.debug(msg)
            elif level == "info":
                logger.info(msg)
            elif level == "warning":
                logger.warning(msg)
            elif level == "error":
                logger.error(msg)
            elif level == "critical":
                logger.critical(msg)

        alert_text = None
        try:
            return func(self, *args, **kwargs)
        except UnexpectedAlertPresentException:
            # Some websites create an alert to communicate an error message instead of throwing the normal exception
            # or instead of things normal people use popups for.
            # However, selenium can be "too quick" to dismiss the alerts and crash.
            alert = Alert(self.driver)
            try:
                alert_text = alert.text
                alert.dismiss()
            except NoAlertPresentException:
                pass
        except JavascriptException as e:
            logger.error("Exception thrown in the browser's Javascript engine.")
            logger.debug(e)
        finally:
            log_records = []
            try:
                log_records = self.driver.get_log("browser")
            except (WebDriverException, WindowClosedError):
                logger.warning("Failed to fetch log records from the driver")

            # Print browser logs
            for record in sorted(log_records, key=lambda x: int(x["timestamp"])):
                record["timestamp"] = datetime.fromtimestamp(record["timestamp"] / 1000).strftime("%H:%M")
                log_line = "[{level}] {message}".format(**record)
                log_level = record["level"]
                to_logger(log_line, log_level)

            # Then alert text, if there was any
            if alert_text is not None:
                logger.error(f"  JS Alert text:  {alert_text}")

    return wrapper


class JavascriptWrapper:
    """
    Helper class for executing built-in javascript scripts or custom
    files and snippets.
    """

    def __init__(self, driver: WebDriver, config: Config = None):
        assert driver
        config = config or Config.default()
        self.driver = driver
        self.config = config

    def save_mhtml(self, filename: str):
        """
        Executes the MHTML saving extension. Saves to the path specified in config.scraping.temp_path.
        Note: If the file already exists, it will not be overwritten.
        """
        self.execute_script('window.postMessage({type: "SAVE_MHTML", filename: "' + filename + '"})')

    def get_full_height(self) -> int:
        """
        Get the full page height, i.e. the height of the document.
        :return: document height in pixels
        """
        return self.execute_file(Path("get_full_page_height.js"))

    def find_iframe_name(self, identifier: str) -> str:
        """
        Looks for an iframe where name, ID, or class equals the identifier,
        and returns its name. Returns empty string if no matching object was found.
        :return: iframe name or empty string
        """
        return self.execute_file(Path("find_iframe_name.js"), identifier)

    def find_viewport(self) -> Rectangle:
        """
        Get the width of the web browser window with content.
        :return: viewport height in pixels
        """
        result = self.execute_file(Path("find_viewport.js"))

        if not result:
            logger.error("Failed to compute get viewport size")
            return Rectangle.empty()

        return Rectangle.from_list([result["x"], result["y"], result["x"] + result["w"], result["y"] + result["h"]])

    def get_element_metadata(self) -> List[Dict[str, Any]]:
        """
        Collects metadata about web page DOM elements: their tags, some of the HTML attributes, position and size on the
        page, CSS styles and classes, inner text.

        Each element on the page is assigned a unique within the scope of the page ``wtl_uid`` and
        has a pointer to the parent DOM element in the ``wtl_parent_uid`` field and are not to be confused
        with ``id`` attribute in HTML which is neither unique nor mandatory.

        :return: a list of JSON objects (in their Python dict form) with HTML attributes, additionally calculated
                properties and unique IDs. Refer to the script code for the keys' names.
        """
        return self.execute_file(Path("get_element_metadata.js"))

    def hide_position_fixed_elements(self, elements: List[str] = None) -> dict:
        """
        Hides page elements that are fixed or sticky (the ones that stay on the page when you scroll)
        by setting their visibility to "hidden".

        Returns a map from element ids (wtl-uid) to the old visibility values.

        Mutates the web page.
        """
        elements = elements or []
        return self.execute_file(Path("hide_position_fixed_elements.js"), elements)

    def show_position_fixed_elements(self, id_to_visibility: dict):
        """
        Set the specified visibility to the elements with ids listed in ``id_to_visibility``.

        The 2nd parameter is expected to be the (possibly accumulated) output of hide_position_fixed_elements.

        Mutates the web page (hopefully undoing the changes made by hide_position_fixed_elements)
        """

        self.execute_file(Path("show_position_fixed_elements.js"), id_to_visibility)

    def element_exists(self, selector: Selector) -> bool:
        """
        Returns True if an element exists, otherwise False.
        """

        return self.execute_file(Path("element_exists.js"), selector.css)

    def disable_animations(self):
        """
        Turns off animation on the page. Works for jQuery by setting a certain flag and for CSS animations by injecting
        an additional style into the page code.

        Mutates the web page.
        """
        self.execute_file(Path("disable_animations.js"))

    def is_page_loaded(self, *_) -> bool:
        """
        Applies some heuristics to check if the page is loaded. But since it is in general a hard question to answer,
        is known to be faulty in some cases.
        """
        # Ignores any input arguments given by WebDriverWait.until()
        return self.execute_file(Path("is_page_loaded.js"))

    def find_active_elements(self) -> list:
        """
        Uses a couple of heuristics to try and find all clickable elements in the page.
        """
        return self.execute_file(Path("find_active_elements.js"))

    def scroll_to(self, x: float, y: float):
        """
        Scroll the page to given coordinates.
        """
        self.execute_script(f"scrollTo({int(x)}, {int(y)});")

    def make_canvas(self):
        """
        Create viewport and page canvases and add them to the DOM.
        """
        self.execute_file(Path("make_canvas.js"))

    def annotate(
        self,
        location: Point,
        color: Color,
        size: int,
        text: str,
        background: Color = Color(0, 0, 0, 0),
        viewport: bool = False,
    ):
        """
        Writes text with a given color on the page.
        Shares an HTML canvas with `highlight`.
        """
        self.make_canvas()
        self.execute_file(
            Path("annotate.js"),
            location.x,
            location.y,
            color.to_str(),
            size,
            text,
            background.to_str(),
            background.a / 255,
            viewport,
        )

    def highlight(self, selector: Selector, color: Color, fill: bool = False, viewport: bool = False):
        """
        Highlight an element as found by a given selector with arbitrary
        color and intensity.
        Note: If more elements can be found, only one will be highlighted.
        Shares an HTML canvas with `annotate`.
        """
        self.make_canvas()
        self.execute_file(Path("highlight.js"), selector.css, color.to_str(), color.a / 255, fill, viewport)

    def clear_highlights(self, viewport: bool = False):
        """Removes all highlights created by :func:`highlight`."""
        self.execute_file(Path("clear_highlights.js"), viewport)

    def click_element(self, selector: Selector):
        """
        Clicks an element found by the given selector.
        Note: If more elements can be found, only one will be clicked.
        """
        self.execute_file([Path("dom.js"), Path("click_element.js")], selector.css)

    def delete_element(self, selector: Selector):
        """
        Deletes an element found by the given selector.
        Note: If more elements can be found, only one will be clicked.
        """
        self.execute_file(Path("delete_element.js"), selector.css)

    def fill_text(self, selector: Selector, value: str):
        """
        Fills an element as found by a given selector with given text.
        Note: If more elements can be found, only one will be used.
        """
        self.execute_file([Path("dom.js"), Path("fill_text.js")], selector.css, value)

    def select(self, selector: Selector, value: str):
        """
        Select an element of a dropdown (select) element.
        Note: If more elements can be found, only one will be used.
        """
        self.execute_file([Path("dom.js"), Path("select.js")], selector.css, value)

    @classmethod
    @functools.lru_cache(maxsize=32)
    def assemble_script(cls, filenames: Iterable[Path]) -> str:
        """
        Concatenates the contents of several Javascript files into one, with caching.
        :param filenames: Path to the JS files either in webtraversallibrary/js or an absolute path.
        """
        contents = []

        for filename in filenames:
            if not filename.exists():
                filename = Path(os.path.dirname(__file__)) / "js" / filename

            assert filename.exists() and filename.is_file()

            with open(filename, "rt") as f:
                contents.append(f.read())

        return "\n".join(contents)

    def execute_file(self, filename: Union[Path, Iterable[Path]], *args, execute_async: bool = False) -> Any:
        """
        Execute the JavaScript code in given file and return the result
        :param filename: Path to the JS file(s) either in webtraversallibrary/js or an absolute path.
        :param execute_async: if True, will wait until the javascript code has called
        arguments[arguments.length-1] and will return its input arguments.
        """
        if isinstance(filename, Path):
            filename = [filename]

        script = JavascriptWrapper.assemble_script(tuple(filename))

        if execute_async:
            return self.execute_script_async(script, *args)
        return self.execute_script(script, *args)

    @safe_selenium_method
    def execute_script(self, script: str, *args) -> Any:
        """
        Execute the JavaScript code in `script` and return the result
        :param script: path to the JS file relative to this package
        """
        return self.driver.execute_script(script, *args)

    @safe_selenium_method
    def execute_script_async(self, script: str, *args) -> Any:
        """
        Execute the JavaScript code in `script` asynchronously and returns the result
        :param script: path to the JS file relative to this package
        """
        return self.driver.execute_async_script(script, *args)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Base representation of the current state of a tab."""

from dataclasses import dataclass, field
from typing import Any, Dict, Set

from .actions import Actions
from .snapshot import PageSnapshot


@dataclass(frozen=True)
class View:
    """
    Base representation of the current state of a tab. Holds a snapshot, a list of
    available actions, and output from the prior classifiers.
    Note: The `metadata` field can be added to arbitrarily, and contents will be deeply
    copied to the next view (do not store too large objects!)
    If you need large metadata storage, use the workflow.metadata instead.
    """

    name: str
    snapshot: PageSnapshot = field(hash=False)
    actions: Actions = field(hash=False, default_factory=Actions, repr=False)
    tags: Set[str] = field(hash=False, default_factory=set)
    metadata: Dict[Any, Any] = field(hash=False, default_factory=dict)

    def copy(self, no_snapshot: bool = False):
        """Creates a shallow copy"""
        return View(
            name=self.name,
            snapshot=None if no_snapshot else self.snapshot,
            actions=None if no_snapshot else self.actions,
            tags=self.tags,
            metadata=self.metadata,
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
The Workflow is the main entry point for using the Web Traversal Library.
"""
from __future__ import annotations

import contextlib
import itertools
import logging
import os
from dataclasses import replace
from pathlib import Path
from time import sleep
from typing import Any, Callable, Dict, List, Union

from selenium import webdriver

from .actions import Abort, Action, Actions, ElementAction, Navigate, Refresh, Revert, Wait
from .classifiers import Classifier
from .color import Color
from .config import Config
from .error import Error
from .geometry import Point, Rectangle
from .goals import FOREVER
from .helpers import ClassifierCollection, FrameSwitcher, MonkeyPatches
from .javascript import JavascriptWrapper
from .logging import setup_logging
from .processtools import TimeoutContext
from .scraper import Scraper
from .selector import Selector
from .snapshot import PageElement, PageSnapshot
from .view import View
from .window import Window

logger = logging.getLogger("wtl")


class Workflow:
    """
    The Workflow is the main entry point for using the Web Traversal Library.
    It will handle setup and teardown of all helper classes and appropriate use of the configuration object.
    Note that you can "run a workflow" manually by creating Window and Scraper objects manually,
    however this is not recommended for general use.
    """

    SINGLE_TAB: str = "tab"

    def __init__(
        self,
        url: Union[str, Dict[str, str], Dict[str, Dict[str, str]]],
        policy: Callable,
        output: Path = None,
        config: Config = None,
        goal: Callable = FOREVER,
        classifiers: List[Classifier] = None,
        patches: Dict[Selector, str] = None,
    ):
        """
        Create a Workflow and reset it.
        :param url: Single string, dict mapping tab name to URL, or dict mapping window name to tab: url dict. Each
        tab name must be unique, even among different windows!
        :param policy: Function taking workflow and views, returning an action or a list of actions per tab/view.
        :param output: Path for storing local data (if needed).
        :param config: Configuration object for this instance.
        :param goal: Called before each policy call, will halt the workflow if it returns True.
        :param classifiers: List of classifiers to run on every snapshots.
        :param patches: A dictionary of selectors to monkeypatch to other destinations.
        """
        setup_logging(log_dir=output if config.debug.save else None)
        config = config or Config.default()
        config.validate()

        self.loop_idx = -1
        self.output = output
        self.config = config
        self._windows: Dict[str, Window] = {}
        self.policy = policy() if policy.__name__ == "__wtl_wrapped" else policy
        self.goal = goal if goal else FOREVER
        self.preload_callbacks: List[Path] = []
        self.postload_callbacks: List[Callable] = []
        self._history: Dict[str, List[View]] = {}
        self._current_tab: str = None
        self._current_window: Window = None
        self._has_quit = False
        self._tabs_cache: List[str] = None
        self.metadata: Dict[Any, Any] = {}
        self.monkeypatches = MonkeyPatches(patches)
        self.classifiers = ClassifierCollection(classifiers)
        self.previous_policy_result = None

        # Basic error handling
        assert self.policy, "Workflow created without a policy!"
        if self.config.debug.save:
            assert self.output, "Saving debug output requires specifying an output path!"

        # Setup starting points
        if isinstance(url, str):
            # Convert url type from str to Dict[str, str]
            url = {Workflow.SINGLE_TAB: url}
        if isinstance(list(url.values())[0], str):
            # Convert url type from Dict[str, str] to Dict[str, Dict[str, str]]]
            url = {Workflow.SINGLE_TAB: url}  # type: ignore
        self._starting_url: Dict[str, Dict[str, str]] = url  # type: ignore

        self.reset()

    def __enter__(self):
        return self

    def __exit__(self, *_, **__):
        """Calls :func:`quit` unless debug.preserve_window in config is True."""
        if not self.config.debug.preserve_window:
            self.quit()

    @Error.wrapped
    def run(self):
        """
        Runs the workflow loop!
        Intializes the tabs to the starting URLs and then
        calls :func:`run_once` in a loop until the goal function returns True
        or all tabs have stopped.
        """
        assert not self._has_quit
        if self.classifiers:
            self.preload_callbacks.append(Path("js/intercept_event_listeners.js"))

        try:
            ctx = TimeoutContext(n_seconds=self.config.timeout) if self.config.timeout else contextlib.nullcontext()
            with ctx:
                while self.success:
                    if self.run_once():
                        break

        except TimeoutError:
            logger.error("Workflow.run() timed out!")

        logger.debug("Workflow.run() completed!")

    @Error.wrapped
    def run_once(self):
        """
        Runs a single iteration of the WTL flow, i.e. snapshots, runs classifiers,
        checks goal, computes actions, calls policy, executes actions.
        Check `loop_idx` member attribute for number of iterations.

        .. note::
            This does not initialize the tabs to their starting URLs. Normally, use :func:`run` instead.

        :return: The boolean output from the goal function.
        """
        self._populate_tabs_cache()

        # Perform required snapshotting
        self.loop_idx += 1
        all_views = self._get_new_views()

        # Stop iterating if we've reached our goal
        goal_result = self.goal(self, all_views)
        if isinstance(goal_result, dict):
            goal_result = goal_result and all(goal_result.values())

        if not goal_result:
            # Query the policy
            try:
                policy_result = self.policy(self, all_views)
                self.previous_policy_result = policy_result
                policy_stopped = False
            except StopIteration:
                logger.info("Received StopIteration from policy, will not continue.")
                policy_stopped = True

            # Execute actions
            if not policy_stopped:
                self._execute_policy_result(policy_result)

        # Save snapshot to disk
        if self.config.debug.save:
            for tab in self.tabs:
                self.tab(tab)
                try:
                    if self.latest_view.snapshot:
                        self.latest_view.snapshot.save(self.output_path)
                except IndexError:
                    pass

        # Stop iterating if goal returned True
        if goal_result:
            return True

        # Abort all remaining tabs if policy is finished
        if policy_stopped:
            self._execute_policy_result({tab: Abort() for tab in self.open_tabs})

        # Reduce memory usage if keeping a lot of views
        if self.config.scraping.history and not self.config.scraping.full_history:
            for tab in self.open_tabs:
                self.tab(tab)
                self.history[-1] = self.history[-1].copy(no_snapshot=True)

        # Continue if there is still something in the policy
        return policy_stopped

    def _get_new_views(self) -> Dict[str, View]:
        all_views: Dict[str, View] = {tab: None for tab in self.tabs}

        for tab in self.open_tabs:
            self.tab(tab)

            # Navigate if neccessary
            initial_url = self.current_window.navigation
            initial_action = None
            if initial_url:
                initial_action = Navigate(initial_url)
                self._perform_action(initial_action)

            # Optimization: Only scrape tab if enforced, action taken, or never scraped before
            if (
                not self.config.scraping.all
                and self.previous_policy_result
                and tab not in self.previous_policy_result
                and self.latest_view
                and self.latest_view.snapshot
            ):
                self.history.append(self.latest_view)
                all_views[tab] = self.history[-1]
                continue

            # Store View
            all_views[tab] = self._get_new_view(tab, initial_action)

        return all_views

    def _get_new_view(self, name: str, initial_action: Action) -> View:
        # Ensure page is fully loaded
        self.current_window.scraper.wait_until_loaded()

        # Run postload callbacks
        for cb in self.current_window.scraper.postload_callbacks:
            cb()

        # Scrape the page
        snapshot = self.current_window.scraper.scrape_current_page()

        # Assemble basic list of actions
        action_list: List[Action] = [Abort(), Refresh(), Navigate(), Wait()]
        action_list += [Revert(step) for step in range(len(self.history))]
        actions = Actions(action_list)

        # Setup metadata
        metadata: Dict[str, Any] = {}
        if self.config.scraping.history:
            if self.history and "next_action" in self.history[self.loop_idx - 1].metadata:
                metadata = self.history[self.loop_idx - 1].metadata.copy()
                metadata["previous_action"] = metadata["next_action"]
            else:
                metadata["previous_action"] = [initial_action]
        metadata["next_action"] = None

        # Create view
        view = View(name=name, snapshot=snapshot, actions=actions, metadata=metadata)

        # Add state to history
        if self.loop_idx < len(self.history):
            self.history[self.loop_idx] = view
        else:
            self.history.append(view)

        # Maintain only one level of history if required
        ct = self.current_tab
        if not self.config.scraping.history:
            for i in range(len(self._history[ct]) - 1):
                self._history[ct][i] = None

        # Run element classifiers
        view.actions.extend(self._run_element_classifiers(snapshot))

        # Run page classifiers
        for classifier in self.classifiers.active_view_classifiers:
            result = classifier.callback(view)
            view.tags.update(result)

        return view

    def _execute_policy_result(self, policy_result: Dict[str, Union[Action, List[Action]]]):
        # If a reset action was given, perform it directly
        revert_actions = [a for a in policy_result.values() if isinstance(a, Revert)]

        revert_actions.sort(key=lambda a: a.view_index)
        if revert_actions:
            self._perform_action(revert_actions[0])
            return

        # Execute other actions in the policy result
        for tab in self.open_tabs:
            # Find corresponding actions
            if tab in policy_result:
                actions = policy_result[tab]
            else:
                keys = [r for r in policy_result if isinstance(r, View) and r.name == tab]
                if not keys:
                    logger.info(f"No action given for {tab}.")
                    actions = []
                else:
                    actions = policy_result[keys[0]]

            # Execute actions
            if not isinstance(actions, list):
                actions = [actions]

            if actions:
                self.tab(tab)

            for i, action in enumerate(actions):
                sleep(self.config.scraping.wait_action)

                if (
                    isinstance(action, ElementAction)
                    and isinstance(action.target, Selector)
                    and not action.target.iframe
                ):
                    action = action.transformed_to_element(self.latest_view.snapshot.elements)
                    actions[i] = action
                self._perform_action(action)

            self.latest_view.metadata["next_action"] = actions

    def create_window(self, name: str) -> Window:
        """Opens a new browser window and adds it to this workflow. Returns the new window."""
        window = Window(
            config=self.config, preload_callbacks=self.preload_callbacks, postload_callbacks=self.postload_callbacks
        )
        self._windows[name] = window
        return window

    @property
    def current_tab(self) -> str:
        """Returns the name of the current tab"""
        return self._current_tab

    @property
    def current_window(self) -> Window:
        """Returns the window object for the current tab."""
        return self._current_window

    @property
    def success(self) -> bool:
        """Returns True if there are any open (i.e. not-cancelled) tabs."""
        return any(window.open_tabs for window in self.windows)

    @property
    def tabs(self) -> List[str]:
        """Returns a list of all tab names."""
        return self._tabs_cache

    @property
    def open_tabs(self):
        """Returns a list of all open tab names."""
        return list(itertools.chain.from_iterable(window.open_tabs for window in self.windows))

    @property
    def windows(self):
        """Returns a list of all window instances."""
        return list(self._windows.values())

    @property
    def history(self) -> List[View]:
        """
        Returns a history of views for the current tab.
        The view stores previous_action and next_action in its metadata for future
        resurrection of the workflow.
        """
        ct = self.current_tab
        if ct not in self._history:
            self._history[ct] = []
            for _ in range(self.loop_idx + 1):
                self._history[ct].append(View(name=ct, snapshot=None))
        return self._history[ct]

    @property
    def aborted(self) -> bool:
        """Returns True if the current tab has been closed."""
        return self.current_window.is_closed(self.current_tab)

    @property
    def js(self) -> JavascriptWrapper:
        """Returns a :class:`JavascriptWrapper` associated to the current window."""
        return self.current_window.js

    @property
    def driver(self) -> webdriver:
        """Returns the WebDriver instance associated to the current window."""
        return self.current_window.driver

    @property
    def scraper(self) -> Scraper:
        """Provides a :class:`Scraper` instance associated to the current window."""
        return self.current_window.scraper

    @property
    def latest_view(self) -> View:
        """Returns the latets :class:`View` taken from the current tab."""
        return self.history[self.loop_idx]

    @property
    def output_path(self) -> Path:
        """Returns path to output directory for current tab and iteration"""
        assert self.output, "No output path was specified!"
        output_path = self.output / str(self.loop_idx)
        if len(self.tabs) > 1:
            output_path = output_path / self.current_tab
        return output_path

    @property
    def post_processing_output_path(self) -> Path:
        """
        Returns path to output directory for post processing output.
        The Workflow does not put anything here itself by default, but this is provided as
        convenience to the user.
        """
        assert self.output, "No output path was specified!"
        output_path = self.output / "post"
        os.makedirs(output_path, exist_ok=True)
        return output_path

    def tab(self, name: str) -> Workflow:
        """Sets the current tab to the given name."""
        if len(self.tabs) > 1 and self._current_tab != name:
            logger.info(f"> Setting tab: {name}...")
        self._current_tab = name
        for window in self.windows:
            if self._current_tab in window.tabs:
                self._current_window = window
                break
        self._current_window.set_tab(name)
        return self

    def window(self, name: str):
        """Returns the window instance with the name provided when created with :func:`create_window`."""
        return self._windows[name]

    def view(self, view: View) -> Workflow:
        """Sets the current tab to the one where given view was taken from."""
        return self.tab(view.name)

    def quit(self):
        """Cleans up all windows. Call this after you are done! Do not use again after this."""
        self._has_quit = True
        for window in self.windows:
            window.quit()

    def frame(self, identifier: str) -> FrameSwitcher:
        """
        Returns a context manager for entering and exiting iframes.
        See `FrameSwitcher` for more details.
        """
        return FrameSwitcher(identifier, self.js, self.driver)

    def _perform_action(self, action: Action):
        if not action:
            logger.warning("None given as action")
            return

        logger.debug(f"Next action: {action}")

        try:
            if isinstance(action, ElementAction):
                has_element_handle = isinstance(action.target, PageElement)
                if self.config.debug.autoscroll and has_element_handle:
                    self.smart_scroll_to(action.target.bounds)  # type: ignore

                if self.config.debug.save and self.config.debug.screenshots and has_element_handle:
                    index = 1 + len([s for s in self.latest_view.snapshot.screenshots if s.startswith("action")])
                    name = f"action{index}"
                    assert name not in self.latest_view.snapshot.screenshots

                    scr = self.scraper.capture_screenshot("action")
                    viewport = self.js.find_viewport()
                    scr.highlight(
                        action.target.bounds - viewport.minima,  # type: ignore
                        Color(255, 0, 0),
                        f"Action: {action.__class__.__name__}",
                    )
                    self.latest_view.snapshot.screenshots[name] = scr

                if self.config.debug.live:
                    self.js.highlight(action.selector, Color.from_str(self.config.debug.action_highlight_color))
                    sleep(self.config.debug.live_delay)

                if has_element_handle:
                    patch = self.monkeypatches.check(action.target.page, action.target)  # type: ignore
                    if patch:
                        action = Navigate(patch)
                        logger.info(f"Action monkeypatched: {action}")

            action.execute(self)
        except NotImplementedError as e:
            logger.error(e)

    def smart_scroll_to(self, element: Rectangle):
        """
        Uses js.scroll_to in a slightly smarter way to minimize
        required scrolls and vertically center elements of interest.
        """
        viewport = self.js.find_viewport()
        y = None

        if element in viewport:
            # Element is completely visible already, do nothing
            pass

        elif element.height > viewport.height:
            # Element is larger than viewport, fit as much as possible
            y = element.y

        elif element.y < viewport.y:
            # Element requires scrolling up, make sure it's inside the screen
            y = element.y

        elif element.y + element.height > viewport.y + viewport.height:
            # Element requires scrolling down, try centering it
            y = element.y + element.height / 2 - viewport.height / 2

        if y is not None:
            self.js.scroll_to(viewport.x, y)
            sleep(self.config.scraping.wait_scroll)

    def _run_element_classifiers(self, snapshot: PageSnapshot) -> List[Action]:
        action_list: List[Action] = []

        for classifier in self.classifiers.active_element_classifiers:
            subset = snapshot.elements.by_score(classifier.subset)
            results = classifier.callback(subset, self)

            if not results:
                continue

            if not isinstance(results, dict):
                results = {"": results}

            for cls_name, cls_result in results.items():
                action_list.extend(self._process_class(classifier, cls_name, cls_result, subset, snapshot))

        return action_list

    def _populate_tabs_cache(self):
        """Updates the cache used by self.tabs property"""
        self._tabs_cache = list(itertools.chain.from_iterable(window.tabs for window in self.windows))

    def _process_class(self, classifier, cls_name, cls_result, subset, snapshot) -> List[Action]:
        # Add the classifier name as the prefix on multi-class predictions
        cls_name = f"{classifier.name}__{cls_name}" if cls_name else classifier.name

        binary_filter = isinstance(cls_result[0], PageElement) if cls_result else True

        if binary_filter:
            cls_result = [(e, 1.0 if e in cls_result else 0.0) for e in subset]

        cls_result.sort(key=lambda x: x[1], reverse=True)

        if binary_filter:
            scaled_result = [value[1] for value in cls_result]
        else:
            scaled_result = classifier.mode.scale([r[1] for r in cls_result])

        cls_result = [(e, r, s) for (e, r), s in zip(cls_result, scaled_result)]

        for element, raw_score, score in cls_result:
            element.raw_scores[cls_name] = raw_score
            element.metadata[cls_name] = classifier.result_type(score)

        if classifier.highlight:
            self._highlight_classifier_result(classifier, cls_name, cls_result, snapshot)

        if classifier.action:
            return [classifier.action(element) for element, _, score in cls_result if score]

        return []

    def reset(self):
        """
        Resets the workflow.
        Does not clear any history.
        """
        assert self.config.scraping.history or self.loop_idx == -1, "Cannot reset if config.scraping.history is False!"

        self.loop_idx = -1
        self.previous_policy_result = None

        # Clear any existing windows
        for window in self.windows:
            if window:
                window.quit()
        self._windows.clear()

        # Create all windows and tabs
        for window_name, tab_names in self._starting_url.items():
            window = self.create_window(window_name)
            for tab_name, url in tab_names.items():
                window.create_tab(tab_name, url=url)
        self._populate_tabs_cache()
        self.tab(self.tabs[0])

    def reset_to(self, view_index: int):
        """
        Resets the Workflow and replays the first ``view_index`` actions.
        History in memory will be mutated.
        Because this resets the `loop_idx` variable to match, saved output will override
        previous output.
        """
        assert self.config.scraping.history, "Cannot reset if config.scraping.history set to False!"
        assert 0 <= view_index < self.loop_idx, "Cannot revert into the future or before the beginning of time!"
        self.reset()
        self.loop_idx = view_index

        for tab in self.tabs:
            self._history[tab] = self._history[tab][: view_index + 1]

        for i in range(view_index + 1):
            logger.info(f"Replaying index {i}...")
            policy_result = {tab: self.tab(tab).history[i].metadata["previous_action"] for tab in self.tabs}
            self._execute_policy_result(policy_result)
            self._get_new_views()

    def _highlight_classifier_result(self, classifier, title, result, snapshot):
        # Given a classifier and its output, call highlighting accordingly.
        # Takes a list of tuples (element, raw_score, score).
        # Highlights according to the scaled score.

        do_screenshots = self.config.debug.screenshots
        if do_screenshots:
            scr = snapshot.new_screenshot(name=classifier.name, of="full")

        def _highlight(index):
            element = result[index][0]
            raw_score = result[index][1]
            score = result[index][2]

            if not score:
                return

            color = replace(classifier.highlight_color, a=int(score * 255))
            self.js.highlight(element.selector, color)
            score_str = f"{score:.2f} " + (f"({raw_score:.2f})" if isinstance(raw_score, float) else f"({raw_score})")

            if self.config.debug.live_annotation:
                self.js.annotate(
                    Point(element.bounds.x + 1, element.bounds.y + element.bounds.height + 10),
                    color,
                    10,
                    f"{title}: {score_str}",
                )

            if do_screenshots:
                scr.highlight(element.bounds, color, text=score_str)

        if isinstance(classifier.highlight, bool):
            for index in range(len(result)):
                _highlight(index)
        elif isinstance(classifier.highlight, int):
            for index in range(min(classifier.highlight, len(result))):
                _highlight(index)
        elif isinstance(classifier.highlight, float):
            for index, entry in enumerate(result):
                if entry[2] > classifier.highlight:
                    _highlight(index)
        else:
            logger.error(f"Invalid classifier.highlight value in {classifier}")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Basic 2-dimensional geometric constructs: points, rectangles, etc. """
from __future__ import annotations

from dataclasses import dataclass
from typing import Iterator, Sequence, Tuple, Union


@dataclass(order=True, frozen=True)
class Point:
    """Represents a point in 2-dimensional plane (e.g. image)."""

    x: float
    y: float

    def __add__(self, other: Point) -> Point:
        return Point(self.x + other.x, self.y + other.y)

    def __sub__(self, other: Point) -> Point:
        return Point(self.x - other.x, self.y - other.y)

    def __mul__(self, scalar: float) -> Point:
        return Point(self.x * scalar, self.y * scalar)

    def __iter__(self) -> Iterator[float]:
        return (data for data in (self.x, self.y))

    @staticmethod
    def zero() -> Point:
        return Point(0, 0)


@dataclass(frozen=True)
class Rectangle:
    """Represents a rectangle in a 2-dimensional plane."""

    minima: Point
    maxima: Point

    def __iter__(self) -> Iterator[float]:
        return (data for data in (self.x, self.y, self.width, self.height))

    def __contains__(self, other: Union[Point, Rectangle]) -> bool:
        return self.contains(other)

    def contains(self, other: Union[Point, Rectangle]) -> bool:
        """
        Tests whether the rectangle contains ``other``.

        :return: ``True`` if other contained in the rectangle, ``False`` otherwise.
        """
        if isinstance(other, Point):
            return self.minima.x <= other.x <= self.maxima.x and self.minima.y <= other.y <= self.maxima.y
        if isinstance(other, Rectangle):
            return (
                self.maxima.x >= other.maxima.x
                and self.maxima.y >= other.maxima.y
                and self.minima.x <= other.minima.x
                and self.minima.y <= other.minima.y
            )

        raise TypeError(f"Expected a Rectangle or a Point, not {type(other)}")

    def clip(self, other: Rectangle) -> Rectangle:
        """
        Return a new rectangle generated by clipping this one by the bounds of ``other``.

        Similar to intersection, but clipping non-intersecting rectangles will result in a degenerate rectangle
        located on one of the edges of ``other``

        :return: New, clipped ``Rectangle`` (possibly degenerate)
        """
        return Rectangle.from_list(
            min(max(self.minima.x, other.minima.x), other.maxima.x),
            min(max(self.minima.y, other.minima.y), other.maxima.y),
            min(max(self.maxima.x, other.minima.x), other.maxima.x),
            min(max(self.maxima.y, other.minima.y), other.maxima.y),
        )

    @property
    def x(self) -> float:
        """The x-coordinate of the lower left vertex"""
        return self.minima.x

    @property
    def y(self) -> float:
        """The y-coordinate of the lower-left vertex"""
        return self.minima.y

    @property
    def bounds(self) -> Tuple[float, float, float, float]:
        """Returns min x, min y, max x, max y"""
        return self.minima.x, self.minima.y, self.maxima.x, self.maxima.y

    @property
    def center(self) -> Point:
        """Return the midpoint of the rectangle"""
        return (self.minima + self.maxima) * 0.5

    @property
    def width(self) -> float:
        return self.maxima.x - self.minima.x

    @property
    def height(self) -> float:
        return self.maxima.y - self.minima.y

    @property
    def area(self) -> float:
        return self.width * self.height

    @staticmethod
    def empty() -> Rectangle:
        """Returns a rectangle of zero area at origo."""
        return Rectangle.from_list(0, 0, 0, 0)

    def resized(self, delta: float) -> Rectangle:
        """
        Returns a resized rectangle, shrinked (inflated for ``delta<0``) by ``2*delta`` in width and in height.
        """
        assert 2 * delta >= -min(self.width, self.height), "Operation only defined for delta >= 2*min(width, height)"
        delta_point = Point(delta, delta)
        return Rectangle(self.minima - delta_point, self.maxima + delta_point)

    def __add__(self, vector: Point) -> Rectangle:
        """Returns a copy of the rectangle translated by ``vector`` """
        return Rectangle(self.minima + vector, self.maxima + vector)

    def __sub__(self, vector: Point) -> Rectangle:
        """Returns a copy of the rectangle inversly translated by ``vector`` """
        return Rectangle(self.minima - vector, self.maxima - vector)

    def __repr__(self):
        return "({}), ({})".format(self.minima, self.maxima)

    @staticmethod
    def bounding_box(rectangles: Sequence[Rectangle]) -> Rectangle:
        """Computes the bounding box of ``rectangles`` """
        if not rectangles:
            raise ValueError("Expected a non-empty sequence of rectangles")

        max_x = max(rect.maxima.x for rect in rectangles)
        max_y = max(rect.maxima.y for rect in rectangles)
        min_x = min(rect.minima.x for rect in rectangles)
        min_y = min(rect.minima.y for rect in rectangles)

        return Rectangle.from_list(max_x, max_y, min_x, min_y)

    @staticmethod
    def intersection(rectangles: Sequence[Rectangle]) -> Rectangle:
        """
        Computes the rectangle which is the intersection of a sequence of ``rectangles``. In case the
        intersection is empty, it returns an empty rectangle.
        """
        if not rectangles or len(rectangles) < 2:
            raise ValueError("Expected a sequence of at least two rectangles")

        if Rectangle.empty() in rectangles:
            return Rectangle.empty()

        max_x = min(rect.maxima.x for rect in rectangles)
        max_y = min(rect.maxima.y for rect in rectangles)
        min_x = max(rect.minima.x for rect in rectangles)
        min_y = max(rect.minima.y for rect in rectangles)

        if max_x < min_x or max_y < min_y:
            return Rectangle.empty()

        return Rectangle.from_list(max_x, max_y, min_x, min_y)

    @staticmethod
    def centered_at(center: Point, radius: float) -> Rectangle:
        """
        A new square centered at ``center`` with side length ``radius``. (The technically correct term here is
        `Apothem <https://en.wikipedia.org/wiki/Apothem>`_.)
        """
        if radius <= 0:
            raise ValueError(f"Cannot instantiate rectangle with non-positive side {2 * radius} centered at {center}")

        radius_point = Point(radius, radius)
        return Rectangle(center - radius_point, center + radius_point)

    @staticmethod
    def from_list(*args) -> Rectangle:
        """Converts tuple/list (x1, y1, x2, y2) to a Rectangle"""
        if len(args) == 1:
            args = args[0]
        if len(args) == 4 and all(isinstance(el, (int, float)) for el in args):
            first_x, first_y, second_x, second_y = args
        else:
            raise ValueError("Invalid argument(s) to create rectangle from!")

        minima = Point(min(first_x, second_x), min(first_y, second_y))
        maxima = Point(max(first_x, second_x), max(first_y, second_y))
        return Rectangle(minima, maxima)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Representation of an RGB(A) color."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Tuple


@dataclass(order=True)
class Color:
    """Representation of an 8-bit RGBA color."""

    r: int
    g: int
    b: int
    a: int = 255

    def __post_init__(self):
        assert all(0 <= channel <= 255 for channel in (self.r, self.g, self.b, self.a))

    @staticmethod
    def from_str(color: str) -> Color:
        """Creates a color from a "#RRGGBBAA" representation. Alpha is optional."""
        if color.startswith("#"):
            color = color[1:]
        r = int(color[0:2], 16)
        g = int(color[2:4], 16)
        b = int(color[4:6], 16)
        a = int(color[6:], 16) if len(color) == 8 else 255
        return Color(r, g, b, a)

    def to_str(self, with_alpha: bool = False) -> str:
        """Creates an "#RRGGBBAA" string from this color. Alpha is optional."""

        def _hex(n):
            return hex(n)[2:].zfill(2)

        color = f"#{_hex(self.r)}{_hex(self.g)}{_hex(self.b)}"
        if with_alpha:
            color += f"{hex(self.a)[2:]}"
        return color.upper()

    def to_tuple(self, with_alpha: bool = False) -> Tuple:
        """
        Returns either a 3- or a 4-tuple with the values,
        depending on the value of `with_alpha`.
        """
        if with_alpha:
            return (self.r, self.g, self.b, self.a)
        return (self.r, self.g, self.b)



import re
import setuptools


with open("README.md", "r") as f:
    long_description = f.read()


with open("webtraversallibrary/version.py", "r") as f:
    version = re.search(r"__version__ = \"(.*?)\"", f.read()).group(1)


setuptools.setup(
    name="webtraversallibrary",
    version=version,
    author="Klarna Bank AB",
    author_email="marcus.naslund@klarna.com",
    description="Abstractions of web interactions",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/klarna-incubator/webtraversallibrary",
    python_requires=">=3.7",
    packages=["webtraversallibrary"],
    include_package_data=True,
    install_requires=[
        "beautifulsoup4>=4.8",
        "html5lib>=1.0.1",
        "pillow>=7.1",
        "requests>=2.24",
        "selenium>=3.141",
        "soupsieve>=2.0"
        "tld",
        "urllib3",
        "prodict>=0.8"
    ],
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Contains built-in simple, common policies and helper decorators for building custom policies."""


import random
from functools import wraps

from .workflow import Workflow


def DUMMY(_workflow, view):
    """Never takes any action."""
    return {v: None for v in view}


def RANDOM(_workflow, view):
    """Picks a completely random (available) action for every view."""
    return {v: random.choice(v.actions) for v in view}


def single_tab(func):
    """
    Decorator for simplifying policies or goals when just one tab is in use.
    Simplifies the API, the views (dict) argument is replaced by the single (first) view instance.
    Also, the policy can now just return action (or list of actions) without a dict mapping.
    """

    @wraps(func)
    def wrapper(workflow, view):
        assert len(view.keys()) == 1
        tab = list(view.keys())[0]
        result = func(workflow, view[tab])
        return {tab: result}

    return wrapper


class _PolicyCoroutine:
    """
    Wrapper for writing coroutines/generators that work as policy functions.
    Do not use directly, instead refer to `single_tab_coroutine` and `multi_tab_coroutine` decorators.
    """

    def __init__(self, p, single: bool):
        self.p = p()
        self.p.send(None)
        self.single = single

    def __call__(self, workflow, view):
        if self.single:
            result = self.p.send((workflow, view[Workflow.SINGLE_TAB]))
            return {Workflow.SINGLE_TAB: result}

        return self.p.send((workflow, view))


def single_tab_coroutine(policy):
    """
    Decorator for simplifying policies given by a generator/coroutine on a single tab.
    Interface will be the same as when using @single_tab for normal functions.
    Allows you to yield from the policy and pass the reference to the function/coroutine
    as you would for a normal method, without instantiating an object.
    """

    def __wtl_wrapped():
        return _PolicyCoroutine(policy, single=True)

    return __wtl_wrapped


def multi_tab_coroutine(policy):
    """
    Decorator for simplifying policies given by a generator/coroutine on multiple tabs.
    Allows you to yield from the policy and pass the reference to the function/coroutine
    as you would for a normal method, without instantiating an object.
    """

    def __wtl_wrapped():
        return _PolicyCoroutine(policy, single=False)

    return __wtl_wrapped
